"""Matchers for comparing to functions and classes etc."""

# pylint: disable=too-few-public-methods

from inspect import isclass

from h_matchers.matcher.core import Matcher

__all__ = ["AnyInstanceOf", "AnyFunction"]


class AnyInstanceOf(Matcher):
    """Matches any instance of another class."""

    def __init__(self, klass):
        super().__init__(klass.__name__, lambda other: isinstance(other, klass))


class AnyCallable(Matcher):
    """Matches any callable at all."""

    def __init__(self):
        super().__init__("* any callable *", callable)


class AnyFunction(Matcher):
    """Matches any function, but not classes."""

    def __init__(self):
        super().__init__(
            "* any function *", lambda item: callable(item) and not isclass(item)
        )



"""Matchers for testing collections have specific items."""

# pylint: disable=too-few-public-methods

from h_matchers.exception import NoMatch
from h_matchers.matcher.core import Matcher


class AnyMappableWithItems(Matcher):
    """Matches any container which contains all specified key value pairs."""

    def __init__(self, key_values):
        super().__init__(
            f"* contains {key_values} *",
            lambda other: self._contains_values(other, key_values),
        )

    @classmethod
    def _contains_values(cls, container, key_values):
        if not hasattr(container, "items"):
            return False

        for key, value in key_values.items():
            if key not in container:
                return False

            if container[key] != value:
                return False

        return True


class AnyIterableWithItemsInOrder(Matcher):
    """Matches any item which contains certain elements in order."""

    def __init__(self, items_to_match):
        super().__init__(
            f"* contains {items_to_match} in any order *",
            lambda other: self._contains_in_order(other, items_to_match),
        )

    @classmethod
    def _contains_in_order(cls, container, items_to_match):
        """Check if each item can be found in order in the container.

        :param container: An iterable of items to check over
        :param items_to_match: An iterable of items to try and match
        :return: A boolean indicating whether each item can be matched
        """
        # Ensure we can work with generators
        try:
            container = list(container)
        except TypeError:
            # It's not even iterable
            return False

        last_index = None

        for item in items_to_match:
            try:
                last_index = (
                    container.index(item)
                    if last_index is None
                    else container.index(item, last_index)
                ) + 1
            except ValueError:
                return False

        return True


class AnyIterableWithItems(Matcher):
    """Matches any item which contains certain elements."""

    def __init__(self, items_to_match):
        super().__init__(
            f"* contains {items_to_match} in any order *",
            lambda other: self._contains_in_any_order(other, items_to_match),
        )

    @classmethod
    def _contains_in_any_order(cls, container, items_to_match):
        """Can every item can be uniquely matched to something in container?.

        :param container: An iterable of items to check over
        :param items_to_match: An iterable of items to try and match
        :return: A boolean indicating whether each item can be matched
        """

        try:
            container = list(container)
        except TypeError:
            # Not even an iterable
            return False

        try:
            cls._is_solvable(cls._cross_match(container, items_to_match))
        except NoMatch:
            return False

        return True

    @classmethod
    def _cross_match(cls, container, items_to_match):
        """Find all possible matches as sets of positions.

        Look through each item and list every position in the `container` that
        they match.
        :yields: Sets of positions which match a particular item
        """
        for item in items_to_match:
            constraint_set = set()
            for other_pos, other_item in enumerate(container):
                if item == other_item:
                    constraint_set.add(other_pos)

            yield constraint_set

    @classmethod
    def _is_solvable(cls, constraints):
        """Works out if a set of constraints are mutually incompatible.

        Takes a list of sets of positions and works out if each set can be
        reduced to a single value which doesn't appear in any other set.

        This indicates that we can match each item to a unique item in the
        parent object, rather than having multiple objects matching the same
        thing.

        :param constraints: An iterable of sets of positions
        :raises NoMatch: If the sets are not resolvable
        :return: True if the sets are resolvable
        """
        # Order items by how many matches they have (ascending)
        sorted_constraints = sorted(constraints, key=len)

        if not sorted_constraints:
            # If there are no items to resolve, we have a match!
            return True

        head, tail = sorted_constraints[0], sorted_constraints[1:]

        # Otherwise we look at each possible match for the first item in turn
        for this_match in head:
            # Pick a match for this item, and then ban every other item from
            # matching this one
            try:
                return cls._is_solvable((set(item) - {this_match} for item in tail))
            except NoMatch:
                # Well this branch didn't work out for some reason...
                continue

        # Oh... no branches worked out
        raise NoMatch()



"""A mixin for AnyCollection which lets you specify the size."""

# pylint: disable=too-few-public-methods

from h_matchers.decorator import fluent_entrypoint
from h_matchers.exception import NoMatch


class SizeMixin:
    """Apply and check size constraints."""

    _min_size = None
    _max_size = None

    @fluent_entrypoint
    def of_size(self, exact=None, at_least=None, at_most=None):
        """Limit the size of the list.

        Can be called as an instance or class method.

        :param exact: Specify an exact size
        :param at_least: Specify a minimum size
        :param at_most: Specify a maximum size
        :raises ValueError: If arguments are missing or incompatible
        :return: self - for fluent chaining
        """
        if exact is not None:
            self._min_size = exact
            self._max_size = exact

        elif at_least is None and at_most is None:
            raise ValueError("At least one option should not be None")

        else:
            if at_least is not None and at_most is not None and at_least > at_most:
                raise ValueError("The upper bound must be higher than the lower bound")

            self._min_size = at_least
            self._max_size = at_most

        return self

    def _check_size(self, other, original=None):
        """Run the size check (if any)."""
        if self._min_size and len(other) < self._min_size:
            raise NoMatch("Too small")

        if self._max_size and len(other) > self._max_size:
            raise NoMatch("Too big")

    def _describe_size(self):
        if self._min_size is None and self._max_size is None:
            return

        if self._min_size == self._max_size:
            yield f"of length {self._min_size}"
        elif self._min_size is not None and self._max_size is not None:
            yield f"with length between {self._min_size} and {self._max_size}"

        elif self._min_size is not None:
            yield f"with length > {self._min_size}"
        else:
            yield f"with length < {self._max_size}"



"""A mixin for AnyCollection which adds checking all items with a matcher."""

# pylint: disable=too-few-public-methods

from h_matchers.decorator import fluent_entrypoint
from h_matchers.exception import NoMatch


class ItemMatcherMixin:
    """Check that all items in the object match an example."""

    _item_matcher = None

    @fluent_entrypoint
    def comprised_of(self, item_type):
        """Specify that every item in the iterable should match a single type.

        For example you can specify that the iterable contains only strings, or
        exactly the number '1'.

        Can be called as an instance or class method.

        :param item_type: The type to match again (can be another matcher)
        :return: self - for fluent chaining
        """

        self._item_matcher = item_type
        return self

    def _check_item_matcher(self, other, original):
        """Check to see if all items in the object match a pattern."""
        if not self._item_matcher:
            return

        items = original.keys() if isinstance(original, dict) else other

        for item in items:
            if not item == self._item_matcher:
                raise NoMatch("Item does not match item matcher")

    def _describe_item_matcher(self):
        if self._item_matcher:
            yield f"of items matching {self._item_matcher}"



"""Mixins to extend the capabilities of the AnyCollection matcher."""

from h_matchers.matcher.collection._mixin.contains import ContainsMixin
from h_matchers.matcher.collection._mixin.item_matcher import ItemMatcherMixin
from h_matchers.matcher.collection._mixin.size import SizeMixin
from h_matchers.matcher.collection._mixin.type import TypeMixin



"""The public interface class for comparing with anything."""

from h_matchers.matcher.code import AnyCallable, AnyFunction, AnyInstanceOf
from h_matchers.matcher.collection import AnyCollection, AnyDict, AnyList, AnySet
from h_matchers.matcher.core import Matcher
from h_matchers.matcher.number import AnyInt
from h_matchers.matcher.string import AnyString

__all__ = ["Any"]


class Any(Matcher):
    """Matches anything and provides access to other matchers."""

    # pylint: disable=too-few-public-methods

    string = AnyString
    int = AnyInt

    function = AnyFunction
    callable = AnyCallable
    instance_of = AnyInstanceOf

    iterable = AnyCollection
    list = AnyList
    set = AnySet
    dict = AnyDict

    def __init__(self):
        super().__init__("* anything *", lambda _: True)
